# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Funcs to manage tokens."""

from functools import partial

from flask import current_app, session
from flask_login import current_user
from invenio_db import db

from ..errors import OAuthClientError, OAuthRejectedRequestError, OAuthResponseError
from ..models import RemoteToken
from ..proxies import current_oauthclient


def get_session_next_url(remote_app):
    """Return redirect url stored in session.

    :param remote_app: The remote application.
    :returns: The redirect URL.
    """
    return session.get("%s_%s" % (token_session_key(remote_app), "next_url"))


def set_session_next_url(remote_app, url):
    """Store redirect url in session for security reasons.

    :param remote_app: The remote application.
    :param url: the redirect URL.
    """
    session["%s_%s" % (token_session_key(remote_app), "next_url")] = url


def token_session_key(remote_app):
    """Generate a session key used to store the token for a remote app.

    :param remote_app: The remote application.
    :returns: The session key.
    """
    return "%s_%s" % (current_app.config["OAUTHCLIENT_SESSION_KEY_PREFIX"], remote_app)


def response_token_setter(remote, resp):
    """Extract token from response and set it for the user.

    :param remote: The remote application.
    :param resp: The response.
    :raises invenio_oauthclient.errors.OAuthClientError: If authorization with
        remote service failed.
    :raises invenio_oauthclient.errors.OAuthResponseError: In case of bad
        authorized request.
    :returns: The token.
    """
    if resp is None:
        raise OAuthRejectedRequestError("User rejected request.", remote, resp)
    else:
        if "access_token" in resp:
            return oauth2_token_setter(remote, resp)
        elif "oauth_token" in resp and "oauth_token_secret" in resp:
            return oauth1_token_setter(remote, resp)
        elif "error" in resp:
            # Only OAuth2 specifies how to send error messages
            raise OAuthClientError(
                "Authorization with remote service failed.",
                remote,
                resp,
            )
    raise OAuthResponseError("Bad OAuth authorized request", remote, resp)


def oauth1_token_setter(remote, resp, token_type="", extra_data=None):
    """Set an OAuth1 token.

    :param remote: The remote application.
    :param resp: The response.
    :param token_type: The token type. (Default: ``''``)
    :param extra_data: Extra information. (Default: ``None``)
    :returns: A :class:`invenio_oauthclient.models.RemoteToken` instance.
    """
    return token_setter(
        remote,
        resp["oauth_token"],
        secret=resp["oauth_token_secret"],
        extra_data=extra_data,
        token_type=token_type,
    )


def oauth2_token_setter(remote, resp, token_type="", extra_data=None):
    """Set an OAuth2 token.

    The refresh_token can be used to obtain a new access_token after
    the old one is expired. It is saved in the database for long term use.
    A refresh_token will be present only if `access_type=offline` is included
    in the authorization code request.

    :param remote: The remote application.
    :param resp: The response.
    :param token_type: The token type. (Default: ``''``)
    :param extra_data: Extra information. (Default: ``None``)
    :returns: A :class:`invenio_oauthclient.models.RemoteToken` instance.
    """
    return token_setter(
        remote,
        resp["access_token"],
        secret="",
        token_type=token_type,
        extra_data=extra_data,
    )


def token_setter(remote, token, secret="", token_type="", extra_data=None, user=None):
    """Set token for user.

    :param remote: The remote application.
    :param token: The token to set.
    :param token_type: The token type. (Default: ``''``)
    :param extra_data: Extra information. (Default: ``None``)
    :param user: The user owner of the remote token. If it's not defined,
        the current user is used automatically. (Default: ``None``)
    :returns: A :class:`invenio_oauthclient.models.RemoteToken` instance or
        ``None``.
    """
    session[token_session_key(remote.name)] = (token, secret)
    user = user or current_user

    # Save token if user is not anonymous (user exists but can be not active at
    # this moment)
    if not user.is_anonymous:
        uid = user.id
        cid = remote.consumer_key

        # Check for already existing token
        t = RemoteToken.get(uid, cid, token_type=token_type)

        if t:
            t.update_token(token, secret)
        else:
            t = RemoteToken.create(
                uid, cid, token, secret, token_type=token_type, extra_data=extra_data
            )
        return t
    return None


def token_getter(remote, token=""):
    """Retrieve OAuth access token.

    Used by flask-oauthlib to get the access token when making requests.

    :param remote: The remote application.
    :param token: Type of token to get. Data passed from ``oauth.request()`` to
        identify which token to retrieve. (Default: ``''``)
    :returns: The token.
    """
    session_key = token_session_key(remote.name)

    if session_key not in session and current_user.is_authenticated:
        # Fetch key from token store if user is authenticated, and the key
        # isn't already cached in the session.
        remote_token = RemoteToken.get(
            current_user.get_id(),
            remote.consumer_key,
            token_type=token,
        )

        if remote_token is None:
            return None

        # Store token and secret in session
        session[session_key] = remote_token.token()

    return session.get(session_key, None)


def token_delete(remote, token=""):
    """Remove OAuth access tokens from session.

    :param remote: The remote application.
    :param token: Type of token to get. Data passed from ``oauth.request()``
        to identify which token to retrieve. (Default: ``''``)
    :returns: The token.
    """
    session_key = token_session_key(remote.name)
    return session.pop(session_key, None)


def oauth_logout_handler(sender_app, user=None):
    """Remove all access tokens from session on logout."""
    oauth = current_oauthclient.oauth
    for remote in oauth.remote_apps.values():
        token_delete(remote)
    db.session.commit()


def make_token_getter(remote):
    """Make a token getter for a remote application."""
    return partial(token_getter, remote)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

from functools import partial, wraps

from flask import (
    current_app,
    flash,
    redirect,
    render_template,
    request,
    session,
    url_for,
)
from flask_security.utils import do_flash, get_message
from invenio_db import db
from invenio_i18n import gettext as _
from markupsafe import Markup

from ..errors import (
    AlreadyLinkedError,
    OAuthClientAlreadyAuthorized,
    OAuthClientError,
    OAuthClientMustRedirectLogin,
    OAuthClientMustRedirectSignup,
    OAuthClientTokenNotFound,
    OAuthClientTokenNotSet,
    OAuthClientUnAuthorized,
    OAuthClientUserNotRegistered,
    OAuthClientUserRequiresConfirmation,
    OAuthError,
    OAuthRejectedRequestError,
)
from ..utils import create_registrationform, fill_form
from .authorized import authorized_handler, extra_signup_handler
from .base import base_disconnect_handler
from .decorators import can_extra_signup
from .token import response_token_setter, token_session_key


def _oauth_error_handler(remote, f, *args, **kwargs):
    """Function to handle exceptions."""
    try:
        return f(remote, *args, **kwargs)
    except OAuthClientError as e:
        current_app.logger.warning(e.message, exc_info=True)
        return oauth2_handle_error(e.remote, e.response, e.code, e.uri, e.description)
    except OAuthClientUnAuthorized:
        return current_app.login_manager.unauthorized()
    except AlreadyLinkedError:
        flash(
            _("External service is already linked to another account."),
            category="danger",
        )
        return redirect(url_for("invenio_oauthclient_settings.index"))
    except OAuthRejectedRequestError:
        flash(_("You rejected the authentication request."), category="info")
        return redirect("/")
    except OAuthClientAlreadyAuthorized:
        return redirect("/")
    except OAuthClientTokenNotFound:
        return redirect("/")
    except OAuthClientUserNotRegistered:
        raise OAuthError("Could not create user.", remote)
    except OAuthClientTokenNotSet:
        raise OAuthError("Could not create token for user.", remote)
    except OAuthClientMustRedirectSignup as e:
        return redirect(
            url_for(
                ".signup",
                remote_app=remote.name,
            )
        )
    except OAuthClientMustRedirectLogin as e:
        return redirect(
            url_for(
                ".login",
                remote_app=remote.name,
            )
        )


#
# Error handling decorators
#
def oauth_resp_remote_error_handler(f):
    """Decorator to handle exceptions."""

    @wraps(f)
    def inner(resp, remote, *args, **kwargs):
        # OAuthErrors should not happen, so they are not caught here. Hence
        # they will result in a 500 Internal Server Error which is what we
        # are interested in.
        _f = partial(f, resp)
        return _oauth_error_handler(remote, _f, *args, **kwargs)

    return inner


def oauth_remote_error_handler(f):
    """Decorator to handle exceptions."""

    @wraps(f)
    def inner(remote, *args, **kwargs):
        # OAuthErrors should not happen, so they are not caught here. Hence
        # they will result in a 500 Internal Server Error which is what we
        # are interested in.
        return _oauth_error_handler(remote, f, *args, **kwargs)

    return inner


#
# Handlers
#
@oauth_resp_remote_error_handler
def authorized_default_handler(resp, remote, *args, **kwargs):
    """Store access token in session.

    Default authorized handler.

    :param remote: The remote application.
    :param resp: The response.
    :returns: Redirect response.
    """
    response_token_setter(remote, resp)
    db.session.commit()
    return redirect(url_for("invenio_oauthclient_settings.index"))


@oauth_resp_remote_error_handler
def authorized_signup_handler(resp, remote, *args, **kwargs):
    """Handle sign-in/up functionality.

    :param remote: The remote application.
    :param resp: The response.
    :returns: Redirect response.
    """
    try:
        next_url = authorized_handler(resp, remote, *args, **kwargs)
        # Redirect to next
        if next_url:
            return redirect(next_url)
        return redirect(url_for("invenio_oauthclient_settings.index"))
    except OAuthClientUserRequiresConfirmation as exc:
        do_flash(
            Markup(
                _(
                    f"A confirmation email has already been sent to {exc.user.email}. Didn't receive it? Click <strong><a href=\"{url_for('security.send_confirmation')}\">here</a></strong> to resend it."
                )
            ),
            category="success",
        )
        return redirect("/")


@oauth_remote_error_handler
def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    This default handler will just delete the remote account link. You may
    wish to extend this module to perform clean-up in the remote service
    before removing the link (e.g. removing install webhooks).

    :param remote: The remote application.
    :returns: Redirect response.
    """
    base_disconnect_handler(remote, *args, **kwargs)
    return redirect(url_for("invenio_oauthclient_settings.index"))


@oauth_remote_error_handler
@can_extra_signup
def signup_handler(remote, *args, **kwargs):
    """Handle extra signup information.

    This should be called when the account info from the remote `info` endpoint is
    not enough to register the user (e.g. e-mail missing): it will show the
    registration form, validate it on submission and register the user.

    :param remote: The remote application.
    :returns: Redirect response or the template rendered.
    """
    session_prefix = token_session_key(remote.name)
    account_info = session.get(session_prefix + "_account_info")

    form = create_registrationform(request.form, oauth_remote_app=remote)
    if not form.is_submitted():
        # Show the form when the user is redirected here after `authorized`
        # (GET request), to fill in the missing information (e.g. e-mail)
        return _render_signup_form(remote, form, account_info)
    elif form.is_submitted():
        # Form is submitted (POST request): validate the user input and register
        # the user
        try:
            next_url = extra_signup_handler(remote, form, *args, **kwargs)
            # after validation, if errors occurred, render the form to display
            # the error messages.
            if form.errors:
                return _render_signup_form(remote, form, account_info)
        except OAuthClientUnAuthorized:
            # Redirect the user to login page
            return redirect(url_for("security.login"))
        except OAuthClientUserRequiresConfirmation as exc:
            # Redirect the user after registration (which doesn't include the
            # activation), waiting for user to confirm his email.
            do_flash(*get_message("CONFIRM_REGISTRATION", email=exc.user.email))
            return redirect("/")

        if next_url:
            return redirect(next_url)
        else:
            return redirect("/")


def _render_signup_form(remote, form, account_info):
    """Return the rendered signup form."""
    fill_form(form, account_info["user"])
    remote_app = current_app.config["OAUTHCLIENT_REMOTE_APPS"][remote.name]
    return render_template(
        current_app.config["OAUTHCLIENT_SIGNUP_TEMPLATE"],
        form=form,
        remote=remote,
        app_title=remote_app.get("title", ""),
        app_description=remote_app.get("description", ""),
        app_icon=remote_app.get("icon", None),
    )


def oauth2_handle_error(remote, resp, error_code, error_uri, error_description):
    """Handle errors during exchange of one-time code for an access tokens."""
    flash(_("Authorization with remote service failed."))
    return redirect("/")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

from functools import wraps

from flask import abort, current_app
from flask_login import current_user
from invenio_accounts.models import Role
from invenio_accounts.proxies import current_datastore
from invenio_db import db
from werkzeug.utils import import_string

from ..models import RemoteAccount


def make_handler(f, remote, with_response=True):
    """Make a handler for authorized and disconnect callbacks.

    :param f: Callable or an import path to a callable
    """
    if isinstance(f, str):
        f = import_string(f)

    @wraps(f)
    def inner(*args, **kwargs):
        if with_response:
            return f(args[0], remote, *args[1:], **kwargs)
        else:
            return f(remote, *args, **kwargs)

    return inner


def authorized_handler(f, authorized_response):
    """Handles an OAuth callback.

    As authorized_handler is deprecated in favor of authorized_response,
    it's has to be wrapped with handler which will be executed
    at the proper time.
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        data = authorized_response()
        return f(*((data,) + args), **kwargs)

    return decorated


def require_more_than_one_external_account(f):
    """Require that the user has more than one external account for login.

    If the user only has one linked external account and no means for logging
    in via local credentials, the decorated function won't be executed.
    This decorator is useful for disconnect handlers, to prevent users from
    disconnecting their last potential means of authentication.
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        if current_user.is_anonymous:
            return f(*args, **kwargs)

        local_login_enabled = current_app.config.get(
            "ACCOUNTS_LOCAL_LOGIN_ENABLED", True
        )
        password_set = current_user.password is not None
        local_login_possible = local_login_enabled and password_set

        remote_apps = current_app.config["OAUTHCLIENT_REMOTE_APPS"]
        accounts = RemoteAccount.query.filter_by(user_id=current_user.get_id()).all()

        # find out all of the linked external accounts for the user
        # that are currently configured and not hidden
        consumer_keys = set()
        remote_apps = current_app.config["OAUTHCLIENT_REMOTE_APPS"]
        for name, remote_app in remote_apps.items():
            if not remote_app.get("hide", False):
                consumer_keys.add(name)  # backcompat with v1.5.4
                remote_app_config = current_app.config[remote_app["params"]["app_key"]]
                consumer_keys.add(remote_app_config["consumer_key"])

        linked_accounts = [acc for acc in accounts if acc.client_id in consumer_keys]

        # execute the function only if local login is possible, or
        # there's more than one linked external account
        if local_login_possible or len(linked_accounts) > 1:
            return f(*args, **kwargs)

        else:
            abort(400)

    return decorated


def _role_needs_update(role_obj, new_role_dict):
    """Checks if role needs to be updated."""
    if role_obj.name != new_role_dict.get(
        "name"
    ) or role_obj.description != new_role_dict.get("description"):
        return True
    return False


def create_or_update_roles(groups, persist_every=500):
    """Create/update DB roles based on the groups provided."""
    roles_ids = set()
    processed_count = 0

    for group in groups:
        try:
            current_app.logger.debug(f"Syncing role: {group['name']}")

            existing_role = current_datastore.find_role_by_id(group["id"])
            if existing_role and existing_role.is_managed:
                current_app.logger.exception(
                    f'Error while syncing roles: A managed role with id: {group["id"]} already exists'
                )
                continue

            existing_role_by_name = current_datastore.find_role(group["name"])
            if existing_role_by_name and existing_role_by_name.is_managed:
                current_app.logger.exception(
                    f'Error while syncing roles: A managed role with name: {group["name"]} already exists'
                )
                continue

            with db.session.begin_nested():
                if not existing_role:
                    role = current_datastore.create_role(
                        id=group["id"],
                        name=group["name"],
                        description=group.get("description"),
                        is_managed=False,
                    )
                    roles_ids.add(role.id)
                elif existing_role and _role_needs_update(existing_role, group):
                    role_to_update = Role(
                        id=group["id"],
                        name=group["name"],
                        description=group.get("description"),
                        is_managed=False,
                    )
                    role = current_datastore.update_role(role_to_update)
                    roles_ids.add(role.id)
                else:
                    roles_ids.add(existing_role.id)

            processed_count += 1

            # Commit every `persist_every` iterations
            if processed_count % persist_every == 0:
                current_datastore.commit()

        except Exception as e:
            current_app.logger.error(
                f"Error while syncing roles: {group['name']}. Error: {e}"
            )
            continue

    # Final commit for any remaining uncommitted changes
    current_datastore.commit()
    return roles_ids



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tasks."""

from celery import shared_task

from .handlers.utils import create_or_update_roles


@shared_task
def create_or_update_roles_task(groups):
    """Task to create/update DB roles based on the groups provided."""
    create_or_update_roles(groups)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C)      2021 TU Wien.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that provides OAuth web authorization support."""

from setuptools import setup

setup()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

from flask_login import current_user
from invenio_db import db

from ..errors import OAuthClientUnAuthorized
from ..models import RemoteAccount
from .utils import require_more_than_one_external_account


@require_more_than_one_external_account
def base_disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    This default handler will just delete the remote account link. You may
    wish to extend this module to perform clean-up in the remote service
    before removing the link (e.g. removing install webhooks).

    :param remote: The remote application.
    :returns: Redirect response.
    """
    if not current_user.is_authenticated:
        raise OAuthClientUnAuthorized()

    with db.session.begin_nested():
        account = RemoteAccount.get(
            user_id=current_user.get_id(), client_id=remote.consumer_key
        )
        if account:
            account.delete()

    db.session.commit()
